# This program is written in Python.  In addition, it requires the
# numerical package Numarray or Numeric. Numarray is available from
# http://www.stsci.edu/resources/software_hardware/numarray

# (c) 2004 G.F. Bertsch, B. Sabbey, and M. Uusnakki.  Noncommercial
# is allowed provided the program is cited in publications that
# make use of it.

import sys, os
from string import join
try:
    from Numeric import *
    from LinearAlgebra import *
except ImportError:
    from numarray import *
    from numarray.linear_algebra import *

def sgn(x):
    if x == 0:
        return 0.0
    elif x < 0:
        return -1.0
    else:
        return 1.0

def rVector(A, B, params):
    residues = []
    for i in range(len(A)):
        residues.append( dot(A[i],params) - B[i] )
    return array(residues)

def MinimaxFit(A, B, indexes=None):
    if indexes == None:
        Matrix = A
        Points = B
    else:
        Matrix = take(A, indexes)
        Points = take(B, indexes)   
        
    LSQres = linear_least_squares(Matrix, Points)
    LSQparams = LSQres[0]

    LSQresidues =  matrixmultiply(Matrix, LSQparams)-Points
    LSQsigmas = [sgn(x) for x in LSQresidues]
    epsilon = sum([x**2.0 for x in LSQresidues])\
              /sum([abs(x) for x in LSQresidues])

    MMres = linear_least_squares(Matrix,epsilon*array(LSQsigmas)+Points)
    MMparams = MMres[0]
    residues = matrixmultiply(Matrix, MMparams)-Points
    return [MMparams, residues]

# Ascent Algorithm from "Approximation Theory" - Cheney
def FindOptimalIndexesAscent(A, B, verbose=1,silent=0):
    Crits = range(A.shape[1]+1)

    mres = MinimaxFit(A, B, Crits)
    mmparams = mres[0]
    Sigmas = [[sgn(x)] for x in mres[1]]

    C = inverse(concatenate((Sigmas,take(A,Crits)), axis=1))
    Y = matrixmultiply(C,take(B,Crits))
    Epsilon = abs(Y[0])
    mmparams = Y[1:]
        
    residues = rVector(A, B, mmparams)
    absres = abs(residues)
    rescat = concatenate((transpose([absres]),transpose([residues]),transpose([array(range(len(residues)))])),axis=1).tolist()
    rescat.sort()

    AlphaDat = rescat[-1]
    if (not silent) and verbose: print AlphaDat  #absolute residue, residue, index
    Alpha = int(AlphaDat[2])
    AlphaRow = A[Alpha,:]

    iters = 0

    while abs(AlphaDat[0]-Epsilon) > 1e-6:
        if (not silent) and verbose:
            print "Run Exchange:", "max|res|=",AlphaDat[0],"> Epsilon=",Epsilon

        # Step1
        Mu = sgn(AlphaDat[1])

        # Step2
        Lambda = matrixmultiply(\
           concatenate( (array([[Mu]]) , [AlphaRow]), axis=1),\
           C)
        # Lambda is a row vector

        # Step3
        BetaPicker = [ [ Mu*Lambda[0,i]/C[0,i], i] for i in range(C.shape[1])]
        BetaPicker.sort()
        Beta = BetaPicker[-1][1]

        if (not silent) and verbose:
            print "Beta=",Beta

        # Step 4 and 5
        for k in range(C.shape[0]):
            C[k,Beta]=C[k,Beta]/Lambda[0,Beta]
            
        for k in range(C.shape[0]):
            for j in range(C.shape[1]):
                if j!=Beta:
                    C[k,j]=C[k,j] - Lambda[0,j]*C[k,Beta]

        # Step 6
        Crits[Beta]=Alpha
        
        if (not silent) and verbose:
            print "Run Analysis:","Crits=",Crits
        Y = matrixmultiply(C,take(B,Crits))
        Epsilon = abs(Y[0])
        mmparams = Y[1:]

        residues = rVector(A, B, mmparams)
        absres = abs(residues)
        rescat = concatenate((transpose([absres]),transpose([residues]),transpose([array(range(len(residues)))])),axis=1).tolist()
        rescat.sort()
        
        AlphaDat = rescat[-1]
        if (not silent) and verbose:
            print AlphaDat # absolute residue, residue, index
        Alpha = int(AlphaDat[2])
        AlphaRow = A[Alpha,:]

        iters = iters+1
        if (not silent) and verbose:
            print "Iterations:",iters

    if not silent:print "Ascent Algorithm Total Iterations:",iters
    return Crits

def Minimax(Al, bl, verbose=1, silent=0, OptimalIndexes=None):
    A = array(Al)
    B = array(bl)

    NumParams = A.shape[1]

    if OptimalIndexes == None:
        OptimalIndexes = FindOptimalIndexesAscent(A, B, verbose, silent)

    MMres = MinimaxFit(A, B, OptimalIndexes)
    MMparams = MMres[0]
    MMresidues = MMres[1]

    if not silent:
        print "\nMiniMax solution for the 'optimal' indexes:"
        for i in OptimalIndexes:
            print i,
        print "\nwith Epsilon:", abs(MMresidues[0])
        print MMparams

    residues = rVector(A, B, MMparams)
    absres = abs(residues)
    residues = residues.tolist()
    absres = absres.tolist()

    bigdiv = []
    while len(bigdiv) < len(OptimalIndexes):
        t = max(absres)
        i = absres.index(t)
        r = residues[i]
        row = [i,r]
        bigdiv.append(row)
        absres[i]=-1

    optres = []
    for i in range(len(OptimalIndexes)):
        row = [OptimalIndexes[i],MMresidues[i]]
        optres.append(row)

    bigdiv.sort()
    optres.sort()

    if (not silent) and verbose:
        print "\nlargest magnitude residues, calculated from minimax params"
        for x in bigdiv:
            print x
    
        print "residues of the six 'optimal' indexes"
        for x in optres:
            print x

    success = 1
    
    for i in range(len(optres)):
        if not optres[i][0]==bigdiv[i][0]:
            success = 0
        if not abs(optres[i][1]-bigdiv[i][1]) < 0.1**6:
            success = 0

    if success:
        return [MMparams, residues, OptimalIndexes]

class Nucleus:
    def __init__(self, Etheory, row):
        self.row = row
        self.Etheory = Etheory
        self.Eexp = None
        self.Ecorr = 0.0
    def getTheory(self):
        return self.Etheory + self.Ecorr
    def getExp(self):
        return self.Eexp
    def notComplete(self):
        return self.Eexp == None
    def residual(self, deltaC=None):
        res = self.getExp() - self.getTheory()
        if deltaC == None:
            return res
        return res - dot(deltaC, self.row)
    
def getLines(filename):
    return [line.split() for line in open(filename, 'r').read().splitlines()]

def getNZlines(filename):
    out = []
    for s in getLines(filename):
        if len(s) == 0:
            continue
        try:
            N, Z = int(s[0]), int(s[1])
            rest = [float(x) for x in s[2:]]
            out.append((N,Z,rest))
        except:
            raise '\n\nbad line in file ' + filename + ': ' + join(s)
    return out

if __name__ == '__main__':
    if len(sys.argv) <= 1:
        print 'Supply file containing interaction energy terms on command line.'
        print 'format:\n\tN Z E_theory I1 I2 I3 ...'
        print '\t...'
        sys.exit(1)
        
    Idict = {}
    for N, Z, r in getNZlines(sys.argv[1]):
        Idict[N,Z] = Nucleus(-abs(r[0]), r[1:]) 
            
    if len(sys.argv) > 2:
        exp_name =  sys.argv[2]
    else:
        exp_name = 'exp.tab'

    for N, Z, r in getNZlines(exp_name):
        g = Idict.get((N,Z), None)
        if g:
            g.Eexp = -abs(r[0])
    
                        
    print 'command line:', join(sys.argv)
    print 'current working directory:', os.getcwd()
    print
    
    row_length = None
    for k, v in Idict.items():
        if v.notComplete():
            del Idict[k]
#            print 'no experimental energy for: ', k
        if row_length != None and len(v.row) != row_length:
            raise '\n\nrows with different lengths: ' + k
        row_length = len(v.row)
        
    constraint = identity(row_length)

    constraint = array([
        [0        ,1,0,0,0,0,0,0,0],
        [-0.859607,0,1,0,0,0,0,0,0],
        [0        ,0,0,1,0,0,0,0,0],
        [0        ,0,0,0,1,0,0,0,0],
        [0        ,0,0,0,0,1,0,0,0],
        [-1.77695 ,0,0,0,0,0,1,0,0],
        [0        ,0,0,0,0,0,0,1,0],
        [0        ,0,0,0,0,0,0,0,1]])

    undo_constraint = transpose(constraint)
    I = array([nucleus.row for nucleus in Idict.values()])
    b = array([nucleus.getExp() - nucleus.getTheory() for nucleus in Idict.values()])
    Ic = dot(I, undo_constraint)
    evals, evecs = eigenvectors(dot(transpose(Ic), Ic))
    evecs = [dot(evec, constraint) for evec in evecs]
    evs = [(evals[i], evecs[i]) for i in range(len(evals))]
    evs.sort(lambda x, y : cmp(y[0], x[0]))
    evecs = [ev[1] for ev in evs]
    evals = [ev[0] for ev in evs]

    def rms(deltaC=None):
        sa = 0.0
        count = len(Idict)
        for nucleus in Idict.values():
            sa += nucleus.residual(deltaC)**2.0
        return sqrt(sa/count)

    def printDeltaC(deltaC):
        print 'delta C: ',
        for dc in deltaC:
            print '%f'%dc,
        print
        
    print 'number of nuclei:', len(Idict)
    print 'original r.m.s error: ', rms()
    print
    
    # least squares refit
    bx = array([0.0]*row_length)
    for nucleus in Idict.values():
        for i in range(len(nucleus.row)):
            bx[i] += nucleus.residual()*nucleus.row[i]
            
    kmax = 4   # search space N=4
    deltaC = array([0.0]*row_length)
    for k in range(kmax):
            deltaC += evecs[k]/evals[k]*dot(evecs[k],bx)
    print '%i-vector least squares r.m.s: %7.4f '% (kmax, rms(deltaC))
#    printDeltaC(deltaC)
    print

    # minimax refit
    searchvecs = array(evecs[:kmax])
    constraint2 = transpose(searchvecs)
    vecp, residues, critical_indices = Minimax(dot(I, constraint2), b, silent=1, verbose=0)
        
    deltaC = array([0.0]*row_length)
    for i in range(len(searchvecs)):
            deltaC += searchvecs[i]*vecp[i]

    print '%i-vector minimax r.m.s: %7.4f' % (kmax, rms(deltaC))
#    printDeltaC(deltaC)
    criticals = [Idict.items()[i] for i in critical_indices]
    criticals.sort()
    print '  N   Z   residual\n'+'-'*18
    for [N, Z], c in criticals:
            print '%3i' % N, '%3i' % Z, '  %+7.4f' % c.residual(deltaC)
    print
        
    if len(sys.argv) > 3:
        for filename in sys.argv[3:]:
            for N, Z, r in getNZlines(filename):
                g = Idict.get((N,Z), None)
                if g:
                    for ri in r:
                        g.Ecorr += ri
            
        for k in Idict.keys():
           if Idict[k].Ecorr == 0.0 :
              del Idict[k]
        print len(Idict)
        I = array([nucleus.row for nucleus in Idict.values()])
        b = array([nucleus.getExp() - nucleus.getTheory() for nucleus in Idict.values()])
        vecp, residues, critical_indices = Minimax(dot(I, constraint2), b, silent=1, verbose=0)
        deltaC = array([0.0]*row_length)
        for i in range(len(searchvecs)):
            deltaC += searchvecs[i]*vecp[i]

        print '%i-vector minimax r.m.s: %7.4f' % (kmax, rms(deltaC))
#        printDeltaC(deltaC)
        criticals = [Idict.items()[i] for i in critical_indices]
        criticals.sort()
        print '  N   Z   residual\n'+'-'*18
        for [N, Z], c in criticals:
            print '%3i' % N, '%3i' % Z, '  %+7.4f' % c.residual(deltaC)
        print
